#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['hrefs']

package_data = \
{'': ['*'],
 'hrefs': ['.pytest_cache/*', '.pytest_cache/v/*', '.pytest_cache/v/cache/*']}

install_requires = \
['pydantic>=1.7', 'starlette']

extras_require = \
{'docs': ['sphinx', 'sphinx-rtd-theme'],
 'test': ['pytest>=6.2',
          'hypothesis>=6.24',
          'fastapi',
          'black',
          'pylint',
          'mypy']}

setup(name='hrefs',
      version='0.1',
      description='Reference support for pydantic models',
      author=None,
      author_email='Jaakko Moisio <jaakko@moisio.fi>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )
